module SingleState where

import Extra(pair)
import State

-- This is a refinement of module State, for the common case where
-- the final state is of the same type as the initial state.

--type State0  d s   s' = d -> s -> s'
--type State   d s u s' = d -> s -> (u,s')

type    SState0 d s   = SState d s ()
newtype SState  d s u = SState (d -> s -> (u,s))


-- a classic reader monad
instance Functor (SState d s) where
  fmap f (SState g) = SState (\d s-> let  (u,s') = g d s
                                     in (f u,s'))
instance Monad (SState d s) where
  return x        = SState (\d s-> (u,s))
  (SState a) >> f = SState (\d s-> let (u,s') = a d s
                                       (SState f') = f u
                                   in f' d s')

-- This enables us to use standard 'do' notation in lots of places in the
-- compiler, improving readability without sacrificing correctness.

-- mapS  becomes mapM
-- mapS0 becomes mapM_
